modesearch <- function(p,k,alpha,mu,Sigma,nk,tol)
{
  ## % modes of k comp normal mixture 
  ## % 
  ## % nk is number of random starting points drawn from mixture
  ## % tol is toleranece on log pdf scale for ending mode search, e.g. <=.01
  ## % example call:
  ## %   [m,pm] = modesearch(p,kmax,pi,mu,Sigma,1000,0.0001);
  ## %
  w=rep(0,k)  ## k is number of components
  z=rep(0,p)  ## p is dimension of data
  Z=matrix(0,p,p) 
  Om=Sigma; 
  a=mu; 
  m=c(); 
  pm=c();
  
  for (j in 1:k)
    {
      Om[j,,]=solve(Sigma[j,,])
      a[j,]=Om[j,,]%*%mu[j,] 
    }
  
  ## % start at randomly sampled points ... and centroids ...
  allx=rbind(mu,my_rmvnmix(nk,alpha,mu,Sigma))
  nk=nk+k;
  allpx = rep(0,nk)
  for(i in 1:nk)
    {
      allpx[i] = my_dmvnmix(allx[i,],alpha,mu,Sigma)
    }
  
  
  for(js in 1:nk)
    {
      x=allx[js,] 
      px=allpx[js]; 
      
      ##     take at most 20 steps from each draw 
      h=0; 
      etol=exp(tol);
      eps=1+etol; 
      while(h<=20 & eps>etol)
        {
          y=z; 
          Y=Z;
          for(j in 1:k)
            {
              w[j]=alpha[j]*my_dmvnorm(x,mu[j,],Sigma[j,,]);
              Y=Y+w[j]*Om[j,,];
              y=y+w[j]*a[j,]
            }
          y=solve(Y)%*%y;
          py=my_dmvnmix(y,alpha,mu,Sigma)
          eps=py/px;
          x=y;
          px=py;
          h=h+1; 
        }
      m=rbind(m, t(x));
      pm=c(pm, px)
    }
  allm = m
  allpm = pm
  
  i = which(!duplicated(round(100*log(pm/max(pm)))))
  u = unique(round(100*log(pm/max(pm))))
  pm=pm[i];
  m=m[i,];
  i=sort(pm,decreasing=TRUE,index.return=TRUE)$ix
  pm=pm[i]; 
  m=m[i,]; 
  ## %[ pm' m' ]
  
  ## % comment out the next command to deliver only modes ... otherwise,
  ## % get antimodes as well ..
  
  ## % return
  
  ## % check that we've found modes not antimodes or shoulders .... 
  im=length(pm); 
  Ip=diag(p)
  imodes=c(); 
  for(i in 1:im)
    {
      G=Z; 
      for(j in 1:k)
        {
          eij=m[i,]-mu[j,]
          G=G + alpha[j]*my_dmvnorm(eij,z,Sigma[j,,])*Om[j,,]%*%(Ip-eij%*%t(eij)%*%Om[j,,]);
        }
      if((det(G)>0))
        {
          imodes=c(imodes,i); 
        }
    }
  
  pm=pm[imodes] 
  m=m[imodes,] 
  
  list(m=m,pm=pm,allm=allm,allx=allx)
  
}

mixturehillclimb <- function(allx,alpha,mu,Sigma,tol)
  {
    
    k = length(alpha)
    p = ncol(allx)
    nk = nrow(allx)
    
    
    w=rep(0,k)  ## k is number of components
    z=rep(0,p)  ## p is dimension of data
    Z=matrix(0,p,p) 
    Om=Sigma; 
    a=mu; 
    m=c(); 
    pm=c();
  
    for (j in 1:k)
      {
        Om[j,,]=solve(Sigma[j,,])
        a[j,]=Om[j,,]%*%mu[j,] 
      }

    allpx = rep(0,nk)
    for(i in 1:nk)
      {
        allpx[i] = my_dmvnmix(allx[i,],alpha,mu,Sigma)
      }

    for(js in 1:nk)
      {
        x=allx[js,] 
        px=allpx[js]; 
        
        ##     take at most 20 steps from each draw 
        h=0; 
        etol=exp(tol);
        eps=1+etol; 
        while(h<=20 & eps>etol)
          {
            y=z; 
            Y=Z;
            for(j in 1:k)
              {
                w[j]=alpha[j]*my_dmvnorm(x,mu[j,],Sigma[j,,]);
                Y=Y+w[j]*Om[j,,];
                y=y+w[j]*a[j,]
              }
            y=solve(Y)%*%y;
            py=my_dmvnmix(y,alpha,mu,Sigma)
            eps=py/px;
            x=y;
            px=py;
            h=h+1; 
          }
        m=rbind(m, t(x));
        pm=c(pm, px)
        
      }
    list(m=m,pm=pm)
  }
