#include <R.h>
#include <Rmath.h>
#include "random.h"
#include "linalg.h"
#include "matrix.h"
#include <math.h>
#include "stuff.h"
#include <stdlib.h>

#define LOG_2_PI 1.83787706640935
#define LOG_PI 1.144729885849400

int my_sample(double* p, int n)
{
  int k;
  double *q, s,d;
  q = new_dup_vector(p, n);
  s = dsum(p,n,1);
  if(s!=1.)
    for(k=0;k<n;k++)
      q[k]/=s;

  for(k=1;k<n;k++)
    q[k]+=q[k-1];
  
  //  printVectorR(q,n);

  my_runif(1,0,1,&d);
  
  for(k=0;d>q[k];k++)
    ;
  
  free(q);
  
  return k;
}

/****************************
 * generate random deviates
 ****************************/
/* 
   Using parametrization where nu is the standard degrees of freedom, 
   and dim is the dimension of the matrix.  Thus the pdf is 
   E(W) = nu S.  Assumes that the matrix Schol has the cholesky factor in 
   the lower triangular.  
 */
void my_rwishart(int n, int nu, double** Schol, int dim, double*** x)
{
  int i,j,k,ell;
  double **foo,**chol,**blah;
  foo = new_2ddarray(dim,dim);
  blah = new_2ddarray(dim,dim);
  // copy the lower triangular cholesky factor out
  chol = new_2ddarray(dim,dim);
  zerov(*chol,dim*dim);
  for(i=0;i<dim;i++)
    for(j=0;j<=i;j++)
      chol[i][j] = Schol[i][j];

  for(j=0;j<n;j++)
    {
      zerov(*foo,dim*dim);

      // fill the diagonal with chi sqaures
      for(k=1;k<=dim;k++)
	{
	  my_rchisq(1,nu+1-k,&(foo[k-1][k-1]));
	  foo[k-1][k-1] = sqrt(foo[k-1][k-1]);
	}
      
      // fill the lower triangular with normals
      for(k=0;k<dim;k++)
	for(ell=0;ell<k;ell++)
	  my_rnorm(1,0,1,&(foo[k][ell]));
      
      // now need to compute blah = chol*foo; W=blah*t(blah)
      // LEFT OFF HERE:  to do this, need to implement a dgemm call in linalg.c
      my_matrix_multiply(chol,dim, dim, 0,foo,dim, dim, 0, blah);
      my_matrix_multiply(blah,dim, dim, 0,blah, dim, dim, 1, x[j]);
      // might want to just compute the cholesky as the matrices are generated
    }
  rm_2ddarray(foo);
  rm_2ddarray(blah);
  rm_2ddarray(chol);
}

/*
  Using alternative parametrization here so that E(W) = S/(nu-2)
  This means that nu = d - p +1 where d is the normal degrees of freedom
  Assumes that Sinvchol is the inverse of the expected value, with the cholesky 
  factor in the lower triangular.  
*/
void my_rinvwishart(int n, int nu, double** Sinvchol, int dim, double*** x)
{
  int i,j,k,ell;
  double **foo,**chol,**blah;
  foo = new_2ddarray(dim,dim);
  blah = new_2ddarray(dim,dim);
  // copy the lower triangular cholesky factor out
  chol = new_2ddarray(dim,dim);
  zerov(*chol,dim*dim);
  for(i=0;i<dim;i++)
    for(j=0;j<=i;j++)
      chol[i][j] = Sinvchol[i][j];
  
  // get back the original degrees of freedom
  nu = nu+dim - 1;

  for(j=0;j<n;j++)
    {
      zerov(*foo,dim*dim);

      // fill the diagonal with chi sqaures
      for(k=1;k<=dim;k++)
	{
	  my_rchisq(1,nu+1-k,&(foo[k-1][k-1]));
	  foo[k-1][k-1] = sqrt(foo[k-1][k-1]);
	}
      
      // fill the lower triangular with normals
      for(k=0;k<dim;k++)
	for(ell=0;ell<k;ell++)
	  my_rnorm(1,0,1,&(foo[k][ell]));
      
      // now need to compute blah = chol*foo; W=blah*t(blah)
      // LEFT OFF HERE:  to do this, need to implement a dgemm call in linalg.c
      my_matrix_multiply(chol,dim, dim, 0,foo,dim, dim, 0, blah);
      my_invert_matrix(blah,dim);
      my_matrix_multiply(blah,dim, dim, 1,blah, dim, dim, 0, x[j]);
      // might want to just compute the cholesky as the matrices are generated
    }
  rm_2ddarray(foo);
  rm_2ddarray(blah);
  rm_2ddarray(chol);
}

void my_rchisq(int n, double a, double* x)
{
  int i;
  for(i=0;i<n;i++)
    x[i] = rchisq(a);
}

void my_runif(int n, double a, double b, double* x){
  int i;
  for(i=0;i<n;i++)
    x[i] = runif(a,b);
}

// for some reason, R's gamma distribution is parametrised as (a, 1/b)
// this gives draws from a gamma distribution with shape a and scale b, 
//  such that the expected value is a/b 
void my_rgamma(int n, double a, double b, double* x)
{
  int i;
  double binv;
  binv = 1./b;

  for(i=0;i<n;i++)
    x[i] = rgamma(a,binv);
}

// gives n draws from a beta distribution with shape parameters a and b
void my_rbeta(int n, double a, double b, double *x)
{
  int i;
  for(i=0;i<n;i++)
    x[i] = rbeta(a,b);
}


void my_rbinom(int n, int counts, double p, int* x)
{
  int i;
  for(i=0;i<n;i++)
    x[i] = (int) rbinom((double) counts, p);

}

void my_rmultinom(int n, int counts, double* p, int ncats, int** x){
  int i;
  for(i=0;i<n;i++)
    rmultinom(counts, p,ncats, x[i]);
}

void my_rmultinomstar(int n, int counts, double* p, int ncats, int** badx, int nbad, int** x)
{
  int i,k,good;
  for(i=0;i<n;i++)
    {
      good=0;
      while(!good)
	{
	  rmultinom(counts,p,ncats,x[i]);
	  good=1;
	  for(k=0;k<nbad;k++)
	    if(ivequal(badx[k],x[i],ncats))
	      good=0;
	}
    }
}
 
// gives n draws from an exponential distribution with rate la
void my_rexp(int n, double la, double* x)
{
  int i;
  double linv;
  linv = 1./la;
  for(i=0;i<n;i++)
    x[i] = rexp(linv);  // the function rexp is parametrized in terms of the mean
}

// gives n draws from a geometric distribution with probability of success p
void my_rgeom(int n, double p, int* x)
{
  int i;
  for(i=0;i<n;i++)
    x[i] = (int) rgeom(p);
}

// gives n draws from a poisson distribution with mean la
void my_rpois(int n, double la, int* x){
  int i;
  for(i=0;i<n;i++)
    x[i] = rpois(la);
}

// gives n draws from a normal distribution with mean mu and standard dev sd
void my_rnorm(int n, double mu, double var, double* x){
  int i;
  for(i=0;i<n;i++)
    x[i] = rnorm(mu,sqrt(var));
}

/*
 * generates draws from a multivariate normal distribution
 * assumes cov is row major, with the cholesky factor in the lower triangular
 */
void my_rmvnorm(int n, double* mu, double** cov, int dim, double** x){

  int t,i,j;//,deg;
  double* rn;

  rn = (double*) malloc(sizeof(double)*dim);
  
  for(t=0;t<n;t++)
    {
      my_rnorm(dim, 0., 1., rn);
      for(j=0;j<dim;j++) {
	x[t][j] = 0;
	for(i=0;i<j+1;i++) {
	  x[t][j] += cov[j][i]*rn[i];
	}
	x[t][j] += mu[j];
      }
    }
  free(rn);
}

void my_rmvt(int n, double* mu, double** cov, double nu,int dim, double** x)
{
  int t,i,j;
  double* rn;
  rn = (double*) malloc(sizeof(double)*dim);
  double foo;
  for(t=0;t<n;t++)
    {
      my_rnorm(dim,0,1,rn);
      my_rchisq(1,nu,&foo);
      foo = sqrt(nu/foo);
      for(j=0;j<dim;j++)
	{
	  x[t][j] = 0;
	  for(i=0;i<j+1;i++)
	    {
	      x[t][j]+=cov[j][i]*rn[i]*foo;
	    }
	  x[t][j]+=mu[j];
	}
    }
  free(rn);
}

// gives n draws from an inverse gamma distribution with shape a and scale b
void my_rinvgamma(int n, double a, double b, double* x)
{
  int i;
  double binv;
  binv = 1./b;
  for(i=0;i<n;i++)
    x[i] = 1./rgamma(a,binv);
}

void my_rdirichlet(int n, double* p, int dim, double **x)
{
  int i,j;
  double s;

  for(i=0;i<n;i++)
    {
      s = 0.;
      for(j=0;j<dim;j++)
	{
	  x[i][j] = rexp(p[j]);  // the r function rexp is actually parametrized in terms of the mean
	  s+=x[i][j];
	}
      for(j=0;j<dim;j++)
      	x[i][j]/=s;
    }
  
}


/**************************************
 * calculate probability density functions 
 **************************************/
// assumes that Sigma has the cholesky in the lower triangular, as usual
void my_dmvt(double* x, double* mu,double** Sigma, double nu, int dim, int logspace, double* d)
{
  double log_det_sigma,discrim,mydim;
  mydim = (double) dim;
  double *xx;
  
  xx = new_dup_vector(x,dim);
  log_det_sigma = my_det(Sigma,dim,1);
  
  // xx:=(xx-mu)
  my_daxpy(dim,-1,mu,1,xx,1);

  // xx:= UpperChol(Sigma)^-1 (x-mu)
  my_dtrsv(CblasTrans,dim,Sigma,dim,xx,1);
  
  // discrim:= [UpperChol(Cov)^-1 (x-mu)]'[UpperChol(Cov)^-1 (x-mu)]
  //        := (x-mu)' LowerChol(Cov)^-1 UpperChol(Cov)^-1 (x-mu)
  discrim = my_ddot(dim,xx,1,xx,1);
  
  *d = lgammafn((nu+ mydim)/2)
    -lgammafn(nu/2) 
    -.5*mydim*log(nu)
    -.5*mydim*LOG_PI 
    - .5*log_det_sigma 
    -.5*(nu+mydim)*log(1+discrim/nu);

  if(!logspace)
    *d = exp(*d);
  
  free(xx);
}

void my_dunif(double x, double a, double b, int logspace, double* d)
{
/*   *d = -log(a-b); */
/*   if(!logspace) */
/*     *d = exp(*d); */
  *d = dunif(x,a,b,logspace);
}

void my_dgamma(double x, double a, double b, int logspace, double* d)
{
  *d = dgamma(x,a,1./b,logspace);
}

void my_dbeta(double x, double a, double b, int logspace, double *d)
{
  *d = dbeta(x,a,b,logspace);
}

void my_dbinom(int x, int counts, double p, int logspace, double *d)
{
  *d = dbinom((double) x, (double) counts, p, logspace);
}

void my_dmultinom(int *x, double* p, int ncats, int logspace, double* d)
{
  int k,counts;
  counts = isum(x,ncats,1);
  *d = lgammafn((double) counts+1.);
  for(k=0;k<ncats;k++)
    *d = *d - lgammafn((double) x[k]+1.) + (double) x[k]*log(p[k]);
  if(!logspace)
    *d = exp(*d);
  
}

void my_dmultinomstar(int *x, double* p, int ncats, int** badx, int nbad, int logspace, double* d)
{
  int k;
  double d1,d2,d3;
  
  for(k=0;k<nbad;k++)
    {
      if(ivequal(x,badx[k],ncats))
	{
	  //	  fprintf(stderr,"%d %d %d %d %d is equal to bad vector %d %d %d %d %d\n",x[0],x[1],x[2],x[3],x[4],badx[k][0],badx[k][1],badx[k][2],badx[k][3],badx[k][4]);
	  *d = (logspace ? 1 : 0);
	  return;
	}
    }
  my_dmultinom(x,p,ncats,0,&d1);
  d2 = 0.;
  for(k=0;k<nbad;k++)
    {
      my_dmultinom(badx[k],p,ncats,0,&d3);
      d2+=d3;
    }
  if(logspace)
    *d = log(d1) - log(1-d2);
  else
    *d = d1/(1-d2);
}

void my_dexp(double x, double la, int logspace, double* d)
{
/*   *d = log(la) - (la*x); */
/*   if(!logspace) */
/*     *d = exp(*d); */
  *d = dexp(x, 1./la, logspace);
}

void my_dgeom(int x, double p, int logspace, double *d)
{
  *d = dgeom((double) x, p, logspace);
}

void my_dpois(int x, double la, int logspace, double *d)
{
  *d = dpois((double) x, la, logspace);
}

void my_dnorm(double x, double mu, double var, int logspace, double* d){
  *d = dnorm(x,mu,sqrt(var),logspace);
}

/*
 * calculates the  multivariate normal pdf
 * assumes cov is row major, with the cholesky factor in the lower triangular
 */
void my_dmvnorm(double* x, double* mu, double** cov, int dim, int logspace, double* d){
  
  double log_det_sigma, discrim;
  double *xx;//,*effmu,*effx,**effcov;
  int i;//,effdim, *effparam, t;
  
 /*  effdim = findEffDim(cov,dim); */

/*   if(effdim==dim) /\* handle the case of a proper mvn *\/ */
/*     { */
      // xx := x
      xx = (double*) malloc(sizeof(double) *dim);
      for(i=0; i<dim; i++) xx[i] = x[i];
      
      log_det_sigma = my_det(cov, dim, 1);
      
      // xx := (xx-mu)
      my_daxpy(dim, -1, mu, 1, xx, 1);
      
      // xx:= UpperChol(Cov)^-1 (x-mu)
      my_dtrsv(CblasTrans, dim, cov, dim, xx, 1);
      
      // discrim:= [UpperChol(Cov)^-1 (x-mu)]'[UpperChol(Cov)^-1 (x-mu)]
      //        := (x-mu)' LowerChol(Cov)^-1 UpperChol(Cov)^-1 (x-mu)
      discrim = my_ddot(dim, xx, 1, xx, 1);

      *d = -0.5 * (discrim + log_det_sigma + (dim*LOG_2_PI));
      
      if(!logspace)
	*d = exp(*d);
      
      free(xx);
/*     } */
/*   else if(effdim==0)   */
/*     { /\*handle the case that this is a degenerate distribution C=0 *\/ */
/*       *d = 1.; */
/*       for(i=0;i<dim;i++) */
/* 	if(x[i]!=mu[i]) */
/* 	  *d = 0.; */
/*       if(logspace) */
/* 	*d = log(*d); */
/*     } */
/*   else  */
/*     {  /\* handle the case of a mix of a pointmass and a mvn *\/ */
/*       /\* calculate the pmf part *\/ */
/*       t=1; */
/*       for(i=0;i<dim;i++) */
/* 	if(isDegen(i,cov)) */
/* 	  if(x[i]!=mu[i]) */
/* 	    t=0; */
      
/*       if(t) */
/* 	{ */
/* 	  /\* find effx, effmu, effcov*\/ */
/* 	  effx = new_zero_vector(effdim); */
/* 	  effmu = new_zero_vector(effdim); */
/* 	  effcov = new_2ddarray(effdim,effdim); */
/* 	  effparam = (int*) malloc(sizeof(int)*effdim); */
/* 	  findEffParams(cov,dim,effparam); */
/* 	  for(i=0;i<effdim;i++) */
/* 	    { */
/* 	      effx[i] = x[effparam[i]]; */
/* 	      effmu[i] = mu[effparam[i]]; */
/* 	    } */
/* 	  extractSubMatrix(cov,effdim,effparam,effcov); */
	  
/* 	  my_dmvnorm(effx,effmu,effcov,effdim,logspace,d); */
	  
/* 	  free(effparam); */
/* 	  free(effx); */
/* 	  free(effmu); */
/* 	  rm_2ddarray(effcov); */
/* 	} */
/*       else */
/* 	{ */
/* 	  *d = 0.;   // indicates there is a degeneracy in the covariance matrix */
/* 	  if(logspace) */
/* 	    *d = log(*d); */
/* 	} */
/*     } */
  
}

void my_dinvgamma(double x, double a, double b, int logspace, double *d)
{
  
  *d = a*log(b) - (b/x) - (a+1)*log(x) - lgammafn(a);
  if(!logspace)
    *d = exp(*d);
}

void my_ddirichlet(double* x, double* p, int dim, int logspace, double* d)
{
  double p0;
  int k;
  p0 = dsum(p,dim,1);
  *d = lgammafn(p0);
  for(k=0;k<dim;k++)
    *d = *d - lgammafn(p[k]) + (p[k]-1.)*log(x[k]);
  if(!logspace)
    *d = exp(*d);
}



/*******************************************
 * Provide R interfaces to some of the above
*******************************************/
void R_rwishart(int* n, int* nu, double* S, int* dim, double* x)
{
  GetRNGstate();
  double **myS, ***myx;
  myS = new_2ddarray(*dim,*dim);
  dupv(*myS,S,(*dim)*(*dim));
  my_chol(myS,*dim);
  myx = new_3ddptrs(*n, *dim, *dim, x);
  my_rwishart(*n,*nu,myS,*dim,myx);
  rm_2ddarray(myS);
  rm_3ddptrs(myx,*n);
     
  PutRNGstate();

}

void R_rinvwishart(int* n, int* nu, double* S, int* dim, double* x)
{
  GetRNGstate();
  double **mySinv, ***myx;
  mySinv = new_2ddarray(*dim,*dim);
  dupv(*mySinv,S,(*dim)*(*dim));
  my_invert_matrix(mySinv,*dim);
  my_chol(mySinv,*dim);
  myx = new_3ddptrs(*n, *dim, *dim, x);
  my_rinvwishart(*n,*nu,mySinv,*dim,myx);
  rm_2ddarray(mySinv);
  rm_3ddptrs(myx,*n);
     
  PutRNGstate();

}

void R_rmvnorm(int* n, double* mu, double* cov, int* dim, double* x)
{
  GetRNGstate();
  double **mycov, **myx;
  mycov = new_2ddarray(*dim, *dim);
  dupv(*mycov, cov, (*dim)*(*dim));
  my_chol(mycov,*dim);

  myx = new_2ddptrs(*n, *dim, x);
  my_rmvnorm(*n, mu, mycov, *dim, myx);

  rm_2ddarray(mycov);
  rm_2ddptrs(myx);
  PutRNGstate();
}

void R_rmvt(int* n, double* mu, double* cov, double* nu, int* dim, double* x)
{
  GetRNGstate();
  double **mycov, **myx;
  mycov = new_2ddarray(*dim,*dim);
  dupv(*mycov,cov,(*dim)*(*dim));
  my_chol(mycov,*dim);
  
  myx = new_2ddptrs(*n,*dim,x);
  my_rmvt(*n,mu,mycov,*nu,*dim,myx);

  rm_2ddarray(mycov);
  rm_2ddptrs(myx);
  PutRNGstate();
}

void R_rinvgamma(int* n, double* a, double* b, double* x)
{
  GetRNGstate();
  my_rinvgamma(*n, *a, *b, x);
  PutRNGstate();
}

void R_rdirichlet(int* n, double* p, int* dim, double* x)
{
  double** myx;
  myx = new_2ddptrs(*n, *dim, x);

  GetRNGstate();
  my_rdirichlet(*n, p, *dim, myx);
  PutRNGstate();

  rm_2ddptrs(myx);
}

void R_dmultinom(int *x, double* p, int* ncats, int* logspace, double* d)
{
  my_dmultinom(x, p, *ncats, *logspace, d);
}

void R_dmultinomstar(int *x, double* p, int* ncats, int* badx, int* nbad, int* logspace, double* d)
{
  int** mybadx;
  mybadx = new_2diptrs(*nbad,*ncats,badx);
  
/*   int k,i; */
/*   for(k=0;k<*nbad;k++) */
/*     { */
/*       for(i=0;i<*ncats;i++) */
/* 	fprintf(stderr,"%d ",mybadx[k][i]); */
/*       fprintf(stderr,"\n"); */
/*     } */
	  
  my_dmultinomstar(x,p,*ncats,mybadx,*nbad,*logspace,d);
  rm_2diptrs(mybadx);
}

void R_rmultinomstar(int* n, int* counts, double* p, int* ncats, int* badx, int* nbad, int* x)
{
  int **mybadx, **myx;
  
  mybadx = new_2diptrs(*nbad,*ncats,badx);
  myx = new_2diptrs(*n,*ncats,x);
  GetRNGstate();
  my_rmultinomstar(*n, *counts, p, *ncats, mybadx, *nbad, myx);
  PutRNGstate();
  rm_2diptrs(mybadx);
  rm_2diptrs(myx);
}

void R_dmvnorm(double* x, double* mu, double* cov, int* dim, int* logspace, double* d)
{
  double **mycov;
  mycov = new_2ddarray(*dim, *dim);
  dupv(*mycov, cov, (*dim)*(*dim));
  my_chol(mycov, *dim);

  my_dmvnorm(x, mu, mycov, *dim, *logspace, d);
  rm_2ddarray(mycov);
}

void R_dmvnmixgrid(double* x,int* n,int* dim,double* p, double* mu, double* cov, int* T,double* dens)
{
  double*** mycov;
  mycov = new_3ddarray(*T,*dim,*dim);
  dupv(**mycov,cov,(*T)*(*dim)*(*dim));
  int t;
  for(t=0;t<(*T);t++)
    my_chol(mycov[t],*dim);
  
  double** myx;
  myx = new_2ddptrs(*n,*dim,x);
  double** mymu;
  mymu = new_2ddptrs(*T,*dim,mu);
  
  int i;
  double foo;
  for(i=0;i<*n;i++)
    {
      dens[i] = 0;
      for(t=0;t<*T;t++)
	{
	  my_dmvnorm(myx[i],mymu[t],mycov[t],*dim,1,&foo);
	  dens[i]+=exp(log(p[t])+foo);
	}
    }
  rm_2ddptrs(myx);
  rm_2ddptrs(mymu);
  rm_3ddarray(mycov,*T);
}

void R_dmvtmixgrid(double* x, int* n, int* dim, double* p, double* mu, double* cov, double* nu, int* T, double* dens)
{
  double*** mycov;
  mycov = new_3ddarray(*T,*dim,*dim);
  dupv(**mycov,cov,(*T)*(*dim)*(*dim));
  int t;
  for(t=0;t<(*T);t++)
    my_chol(mycov[t],*dim);
  
  double** myx;
  myx = new_2ddptrs(*n,*dim,x);
  double** mymu;
  mymu = new_2ddptrs(*T,*dim,mu);
  
  int i;
  double foo;
  for(i=0;i<*n;i++)
    {
      dens[i] = 0;
      for(t=0;t<*T;t++)
	{
	  my_dmvt(myx[i],mymu[t],mycov[t],nu[t],*dim,1,&foo);
	  dens[i]+=exp(log(p[t])+foo);
	}
    }
  rm_2ddptrs(myx);
  rm_2ddptrs(mymu);
  rm_3ddarray(mycov,*T);
  
}
void R_dmvt(double* x, double* mu, double* cov, double* nu,int* dim, int* logspace, double* d)
{
  double **mycov;
  mycov = new_2ddarray(*dim,*dim);
  dupv(*mycov,cov,(*dim)*(*dim));
  my_chol(mycov,*dim);

  my_dmvt(x,mu,mycov,*nu,*dim,*logspace,d);
  rm_2ddarray(mycov);

}
void R_dinvgamma(double* x, double* a, double* b, int* logspace, double* d)
{
  my_dinvgamma(*x, *a,*b, *logspace, d);
}

void R_ddirichlet(double* x, double* p, int* dim, int* logspace, double* d)
{
  my_ddirichlet(x, p, *dim, *logspace, d);
}


/* below here was used for testing, but R already has these functions */
/* void R_dunif(double* x, double* a, double* b, int* logspace, double* d) */
/* { */
/*   my_dunif(*x, *a, *b, *logspace, d); */
/* } */

/* void R_dgamma(double* x, double* a, double* b, int* logspace, double* d) */
/* { */
/*   my_dgamma(*x, *a, *b, *logspace, d); */
/* } */

/* void R_dbeta(double* x, double* a, double* b, int* logspace, double *d) */
/* { */
/*   my_dbeta(*x, *a, *b, *logspace, d); */
/* } */

/* void R_dbinom(int* x, int* counts, double* p, int* logspace, double *d) */
/* { */
/*   my_dbinom(*x, *counts, *p, *logspace, d); */
/* } */


/* void R_dexp(double* x, double* la, int* logspace, double *d) */
/* { */
/*   my_dexp(*x, *la, *logspace, d); */
/* } */

/* void R_dgeom(int* x, double* p, int* logspace, double *d) */
/* { */
/*   my_dgeom(*x, *p, *logspace, d); */
/* } */

/* void R_dpois(int* x, double* la, int* logspace, double *d) */
/* { */
/*   my_dpois(*x, *la, *logspace, d); */
/* } */

/* void R_dnorm(double* x, double* mu, double* var, int* logspace, double* d) */
/* { */
/*   my_dnorm(*x, *mu, *var, *logspace, d); */
/* } */

/* void R_runif(int* n, double* a, double* b, double* x) */
/* { */
/*   GetRNGstate(); */
/*   my_runif(*n, *a, *b, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rgamma(int* n, double* a, double *b, double *x) */
/* { */
/*   GetRNGstate(); */
/*   my_rgamma(*n, *a, *b, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rbeta(int* n, double* a, double *b, double *x) */
/* { */
/*   GetRNGstate(); */
/*   my_rbeta(*n, *a, *b, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rbinom(int* n, int* counts, double* p, int* x) */
/* { */
/*   GetRNGstate(); */
/*   my_rbinom(*n, *counts, *p, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rmultinom(int* n, int* counts, double* p, int* ncats, int* x) */
/* { */

/*   int  **myx; */
/*   myx = new_2diptrs(*n, *ncats, x); */

/*   GetRNGstate(); */
/*   my_rmultinom(*n, *counts, p, *ncats, myx); */
/*   PutRNGstate(); */

/*   rm_2diptrs(myx); */
/* } */

/* void R_rexp(int* n, double* la, double* x) */
/* { */
/*   GetRNGstate(); */
/*   my_rexp(*n, *la, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rgeom(int* n, double* p, int* x) */
/* { */
/*   GetRNGstate(); */
/*   my_rgeom(*n, *p, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rpois(int* n, double* la, int* x) */
/* { */
/*   GetRNGstate(); */
/*   my_rpois(*n, *la, x); */
/*   PutRNGstate(); */
/* } */

/* void R_rnorm(int* n, double* mu, double* sd, double* x) */
/* { */
/*   GetRNGstate(); */
/*   my_rnorm(*n, *mu, *sd, x); */
/*   PutRNGstate(); */
/* } */
