#ifndef __random_h__
#define __random_h__

// sample once from a discrete distribution
int my_sample(double* p, int n);

// generate random deviates
void my_rchisq(int n, double a, double* x);
void my_rwishart(int n, int nu, double** Schol, int dim, double*** x);
void my_rinvwishart(int n, int nu, double** Sinvchol, int dim, double*** x);

void my_runif(int n, double a, double b, double* x);
void my_rgamma(int n, double a, double b, double *x);
void my_rbeta(int n, double a, double b, double* x);
void my_rbinom(int n, int counts, double p, int* x);
void my_rmultinom(int n, int counts, double* p, int ncats, int** x);
void my_rexp(int n, double la, double* x);
void my_rgeom(int n, double p, int* x);
void my_rpois(int n, double la, int* x);
void my_rnorm(int n, double mu, double var, double* x);
void my_rmvnorm(int n, double* mu, double** cov, int dim, double** x);
void my_rmvt(int n, double* mu, double** cov, double nu,int dim, double** x);

void my_rinvgamma(int n, double a, double b, double *x);
void my_rdirichlet(int n, double* p, int dim, double **x);
void my_rmultinomstar(int n, int counts, double* p, int ncats, int** badx, int nbad, int** x);

// calculate pdfs
void my_dunif(double x, double a, double b, int logspace, double* d);
void my_dgamma(double x, double a, double b, int logspace, double* d);
void my_dbeta(double x, double a, double b, int logspace, double *d);
void my_dbinom(int x, int counts, double p, int logspace, double *d);
void my_dmultinom(int *x, double* p, int ncats, int logspace, double* d);
void my_dexp(double x, double la, int logspace, double *d);
void my_dgeom(int x, double p, int logspace, double *d);
void my_dpois(int x, double la, int logspace, double *d);
void my_dnorm(double x, double mu, double var, int logspace, double* d);
void my_dmvnorm(double* x, double* mu, double** cov, int dim, int logspace, double *d);
void my_dinvgamma(double x, double a, double b, int logspace, double* d);
void my_ddirichlet(double* x, double* p, int dim, int logspace, double* d);
void my_dmultinomstar(int *x, double* p, int ncats, int** badx, int nbad, int logspace, double* d);
void my_dmvt(double* x, double* mu,double** Sigma, double nu, int dim, int logspace, double* d);

//  functions for R
void R_rmvt(int* n, double* mu, double* cov, double* nu, int* dim, double* x);
void R_dmvtmixgrid(double* x, int* n, int* dim, double* q, double* mu, double* cov, double* nu, int* J, double* dens);
void R_dmvnmixgrid(double* x,int* n,int* dim,double* p, double* mu, double* cov, int* T,double* dens);
void R_dmvt(double* x, double* mu, double* cov, double* nu,int* dim, int* logspace, double* d);
void R_rwishart(int* n, int* nu, double* S, int* dim, double* x);
void R_rinvwishart(int* n, int* nu, double* Sinv, int* dim, double* x);
void R_rmvnorm(int* n, double* mu, double* cov, int* dim, double* x);
void R_rinvgamma(int* n, double* a, double* b, double *x);
void R_rdirichlet(int* n, double* p, int* dim, double *x);

void R_dmultinom(int *x, double* p, int* ncats, int* logspace, double* d);
void R_dmvnorm(double* x, double* mu, double* cov, int* dim, int* logspace, double* d);
void R_dinvgamma(double* x, double* a, double* b, int* logspace, double* d);
void R_ddirichlet(double *x, double *p, int* dim, int* logspace, double* d);

void R_rmultinomstar(int* n, int* counts, double* p, int* ncats, int* badx, int* nbad, int* x);
void R_dmultinomstar(int *x, double* p, int* ncats, int* badx, int* nbad, int* logspace, double* d);

// R already has these - all of these are tested and working properly

/* void R_dunif(double* x, double* a, double* b, int* logspace, double* d); */
/* void R_dgamma(double* x, double* a, double* b, int* logspace, double* d); */
/* void R_dbeta(double* x, double* a, double* b, int* logspace, double *d); */
/* void R_dbinom(int* x, int* counts, double* p, int* logspace, double *d); */
/* void R_dexp(double* x, double* la, int* logspace, double *d); */
/* void R_dgeom(int* x, double* p, int* logspace, double *d); */
/* void R_dpois(int* x, double* la, int* logspace, double *d); */
/* void R_dnorm(double* x, double* mu, double* var, int* logspace, double* d); */

/* void R_runif(int* n, double* a, double* b, double* x); */
/* void R_rgamma(int* n, double* a, double* b, double *x); */
/* void R_rbeta(int* n, double* a, double* b, double* x); */
/* void R_rbinom(int* n, int* counts, double* p, int* x); */
/* void R_rmultinom(int* n, int* counts, double* p, int* ncats, int* x); */
/* void R_rexp(int* n, double* la, double* x); */
/* void R_rgeom(int* n, double* p, int* x); */
/* void R_rpois(int* n, double* la, int* x); */
/* void R_rnorm(int* n, double* mu, double* sd, double* x); */






#endif
